/* License
 * 
 * Copyright 1994-2006 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *  
 *  * Redistribution of source code must retain the above copyright notice,
 *	this list of conditions and the following disclaimer.
 * 
 *  * Redistribution in binary form must reproduce the above copyright notice,
 *	this list of conditions and the following disclaimer in the
 *	documentation and/or other materials provided with the distribution.
 * 
 * Neither the name of Sun Microsystems, Inc. or the names of contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *  
 * This software is provided "AS IS," without a warranty of any kind. ALL
 * EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND WARRANTIES, INCLUDING
 * ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE
 * OR NON-INFRINGEMENT, ARE HEREBY EXCLUDED. SUN MICROSYSTEMS, INC. ("SUN")
 * AND ITS LICENSORS SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY LICENSEE
 * AS A RESULT OF USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE OR ITS
 * DERIVATIVES. IN NO EVENT WILL SUN OR ITS LICENSORS BE LIABLE FOR ANY LOST
 * REVENUE, PROFIT OR DATA, OR FOR DIRECT, INDIRECT, SPECIAL, CONSEQUENTIAL,
 * INCIDENTAL OR PUNITIVE DAMAGES, HOWEVER CAUSED AND REGARDLESS OF THE THEORY
 * OF LIABILITY, ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE,
 * EVEN IF SUN HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 *  
 * You acknowledge that this software is not designed, licensed or intended
 * for use in the design, construction, operation or maintenance of any
 * nuclear facility. 
 */

import java.util.Vector;
import java.util.Random;

/**
* Entity is the base class for objects that move through space through time.
* No state is actually stored on the fields of this class itself.  
* Instead, state is calculated for any given point in time by evaluating
* the plots recorded for this entity.
*/
public class Entity 
{
    public final static int TYPE_ENTITY = 0;
    public final static int TYPE_SHIP = 1;
    public final static int TYPE_SHOT = 2;

    /**
    * Used to indicate a null int value.
    */
    public final static short NIL = Short.MIN_VALUE;
    
    // used to limit the plot history
    private final static byte PLOTS_LIMIT = 50;

    /** 
    * This entity's unique identifier:
    * A combination of the playerid and 
    * the tick of creation followed by
    * the entity's type flags, if any.
    */
    protected long entityID;
    
    /**
    * All state is calculated from this ordered
    * list of plots.
    */ 
    protected Vector plots;
    
    /**
    * Default constructor initializes common state.
    */
    protected Entity()
    {
        plots = new Vector();
        entityID = (long) SpaceWar.playerID << 32;
        //TODO: ensure reasonably unique id
        entityID = entityID | ( 0x0FFF0000 & new Random().nextInt() );
    }
    
    /**
    * Default constructor initializes with specified id.
    */
    public Entity( long inID )
    {
        this();
        entityID = inID;
    }
    
    /**
    * Copy constructor used to create shallow clones for calculation.
    */
    protected Entity( Entity inOriginal )
    {
        this();
        entityID = inOriginal.entityID;
        plots = inOriginal.plots;
    }
    
    /**
    * Returns a shallow clone of this entity.
    */
    public Entity copy()
    {
        return new Entity( this );
    }
    
    /**
    * Returns this entity's unique identifier.
    */
    public long getID()
    {
        return entityID;
    }
    
    
    /**
    * Returns the name of the image resource
    * for any associated sprites.
    * Subclasses should customize.
    */
    public String getImageName()
    {
        return "/ship-12.png";
    }

    /**
    * Returns whether the specified player owns this entity.
    */
    public boolean isOwnedBy( int inID )
    {
        return ( ( ( entityID >> 32 ) ) == inID );
    }

    /**
    * Returns whether this entity has not yet exited the playfield.
    */
    public boolean isAlive( int tick )
    {
        Plot plot = getPositionForTick( tick );
        if ( plot != null && ( plot.type & Plot.TYPE_EXIT ) != 0 ) return false;
        return true;
    }

    /**
    * Returns the size of this entity
    * at the specified tick.
    */
    public int getSize( int tick )
    {
        return 0;
    }

    /**
    * Returns the maximum allowed 
    * rate of change for direction (heading)
    * at the specified tick.
    */ 
    public short getDirectionDelta( int tick )
    {
        return 0;
    }
    
    /**
    * Returns the maximum allowed 
    * rate of change for magnitude (speed)
    * at the specified tick.
    */ 
    public short getMagnitudeDelta( int tick )
    {
        return 0;
    }
    
    /**
    * Returns the maximum allowed 
    * rate of speed at the specified tick.
    */ 
    public short getMagnitudeMax( int tick )
    {
        return 0;
    }
    
    /**
    * Returns the minimum allowed 
    * rate of speed at the specified tick.  
    * May be negative if reverse is allowed.
    */ 
    public short getMagnitudeMin( int tick )
    {
        return 0;
    }
    
    /**
    * Returns a plot containing the position
    * for this entity at the specified point in
    * time, or null if the entity has not yet 
    * entered or has exited the space.
    */
    public Plot getPositionForTick( int inTick )
    {
        Plot next = null;
        Plot prev = null;
        
        // step backwards until we find a resolved plot
        int index = search( plots, inTick );
        if ( index == -1 ) return null; // has not entered space
        
        for ( ; index >= 0; index-- )
        {
            prev = (Plot) plots.elementAt( index );
            if ( prev.x != NIL ) break; // if any are not nil, all are not nil.
        }
        if ( index == -1 ) throw new RuntimeException( "No resolved plots." );
        
        //// find next plot, if any
        next = null;
        if ( plots.size() > ++index ) next = (Plot) plots.elementAt( index );
                
        //System.out.println( "prev: " + prev );
        //System.out.println( "next: " + next );
        
        Plot p = new Plot( prev );

        // if no next plot and no target trajectory
        if ( p.tdir == NIL && p.tmag == NIL )
        { // we can linearly interpolate
            p.x += COS[p.dir*2/45]*p.mag*(inTick-p.tick);
            p.y -= SIN[p.dir*2/45]*p.mag*(inTick-p.tick);
        }
        else // otherwise, we need to step through
        {
            // get current deltas
            short ddir = getDirectionDelta( inTick );
            short dmag = getMagnitudeDelta( inTick );
            
            // copy to local (avoids null checks)
            int tick = p.tick;
            int x = p.x; 
            int y = p.y;
            short dir = p.dir;
            short mag = p.mag;
            short tdir = p.tdir;
            short tmag = p.tmag;

            // make sure target speed is within current allowance
            if ( tmag != NIL ) 
            {
                if ( tmag > getMagnitudeMax( inTick ) ) tmag = getMagnitudeMax( inTick );
                else
                if ( tmag < getMagnitudeMin( inTick ) ) tmag = getMagnitudeMin( inTick );
            }
        
            // calculate forward until we arrive at the desired tick
            while ( ++tick <= inTick )
            {
                // move entity
                x += COS[dir*2/45]*mag;
                y -= SIN[dir*2/45]*mag;
            
                // apply direction delta
                if ( tdir != NIL && dir != tdir && ddir != 0 )
                {
                    int dist = tdir - dir;
                    if ( dist > 180 )
                    {
                        dist = tdir - (dir+360) ;
                    }
                    else
                    if ( dist < -180 )
                    {
                        dist = (tdir+360) - dir;
                    }
                    
                    if ( dist > ddir )
                    {
                        dir = (short) ( dir + ddir );
                    }
                    else
                    if ( dist < -ddir )
                    {
                        dir = (short) ( dir - ddir );
                    }
                    else // within ddir of target
                    {
                        dir = tdir;
                    }
                    
                    dir = (short) ((dir+360)%360);
                }

                // apply magnitude delta
                if ( tmag != NIL && mag != tmag && dmag != 0 )
                {
                    if ( tmag > mag )
                    {
                        mag = (short) ( mag + dmag );
                        if ( mag > tmag ) mag = tmag;
                    }
                    else
                    {
                        mag = (short) ( mag - dmag );
                        if ( mag < tmag ) mag = tmag;
                    }
                }
                
                // if no next plot
                if ( next == null )
                {
                    // if we were changing trajectory
                    if ( tdir != NIL || tmag != NIL )
                    {
                        // if we've reached target trajectory
                        if ( tdir == dir && tmag == mag )
                        {
                            // add a next plot for us to mark as waypoint
                            next = new Plot();
                            next.tick = tick;
                            insertPlot( next );
                            // this gets populated below
                        }
                    }
                }

                // if there's a target plot
                if ( next != null )
                {
                    if ( tick == next.tick ) 
                    { // advance to next plot
                        prev = next;
                        next = null;
                        if ( plots.size() > ++index ) next = (Plot) plots.elementAt( index );

                        // update waypoint
                        prev.x = x;
                        prev.y = y;
                        prev.dir = dir;
                        prev.mag = mag;
                        
                        // update deltas in case they changed
                        ddir = getDirectionDelta( inTick );
                        dmag = getMagnitudeDelta( inTick );
                    }
                }
            }
            
            // copy back
            p.tick = tick;
            p.x = x; 
            p.y = y;
            p.dir = dir;
            p.mag = mag;
            p.tdir = tdir;
            p.tmag = tmag;

        }
        
        // return current position at current time
        p.tick = inTick;
        return p;
    }

    public void insertPlot( Plot inPlot )
    {
        if ( inPlot == null ) { System.err.println( "insertPlot: should never happen" ); return; }
        
        if ( inPlot.tdir != NIL ) inPlot.tdir = (short) (inPlot.tdir%360);
        
        //TODO: need to clear waypoints in all existing plots after this one
        
        //TODO: need to then walk forward again to the current tick from the 
        // plot preceding the added plot
    
        plots.insertElementAt( inPlot, search( plots, inPlot.tick )+1 );
        if ( plots.size() > PLOTS_LIMIT ) plots.removeElementAt( 0 );
    }
    
    /**
    * Returns the index of the plot whose tick is nearest but not over
    * the specified value, or -1 if all are over or none exist.
    */
    private static int search( Vector v, int target )
    { 
        int high = v.size(), low = -1, probe;
        while ( high - low > 1 )
        {
            probe = (high + low) / 2;
            if ( (((Plot)v.elementAt(probe)).tick) > target)
                high = probe;
            else
                low = probe;
        }
        return low;
    }

    /**
    * Returns the range to the specified entity
    * at the specified tick.
    */
    public long getRangeSquared( Entity inEntity, int tick )
    {
        Plot us = getPositionForTick( tick );
        Plot them = inEntity.getPositionForTick( tick );
        
        return ( us.x - them.x ) * ( us.x - them.x ) 
             + ( us.y - them.y ) * ( us.y - them.y );
    }

    /**
    * Returns whether this entity collides with the specified entity
    * at this entity's current tick.
    */
    public boolean collides( Entity inEntity, int tick )
    {
        return ( (getRangeSquared(inEntity, tick) - getSize(tick)*getSize(tick) 
               - inEntity.getSize(tick)*inEntity.getSize(tick)) < 0 );
    }
    
    /**
    * Take damage from a collision with the specified entity
    * at the specified tick.  Return the amount of damage sustained.
    * This implementation returns zero.
    */
    public int collide( Entity inEntity, int tick )
    {
        return 0;
    }

    /**
    * Overridden to always compare by reference.
    * (No need to override hashCode.)
    */
    public final boolean equals( Object inObject )
    {
        return ( inObject == this );
    }
    
    // lookup tables: borrowed from MicroTank

    private static final int[] COS = {
            0,  383,  707,  924,
         1000,  924,  707,  383,
            0, -383, -707, -924,
        -1000, -924, -707, -383
    };

    private static final int[] SIN = {
         1000,  924,  707,  383,
            0, -383, -707, -924,
        -1000, -924, -707, -383,
            0,  383,  707,  924
    };
    
    
}
