/* License
 * 
 * Copyright 1994-2006 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *  
 *  * Redistribution of source code must retain the above copyright notice,
 *	this list of conditions and the following disclaimer.
 * 
 *  * Redistribution in binary form must reproduce the above copyright notice,
 *	this list of conditions and the following disclaimer in the
 *	documentation and/or other materials provided with the distribution.
 * 
 * Neither the name of Sun Microsystems, Inc. or the names of contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *  
 * This software is provided "AS IS," without a warranty of any kind. ALL
 * EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND WARRANTIES, INCLUDING
 * ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE
 * OR NON-INFRINGEMENT, ARE HEREBY EXCLUDED. SUN MICROSYSTEMS, INC. ("SUN")
 * AND ITS LICENSORS SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY LICENSEE
 * AS A RESULT OF USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE OR ITS
 * DERIVATIVES. IN NO EVENT WILL SUN OR ITS LICENSORS BE LIABLE FOR ANY LOST
 * REVENUE, PROFIT OR DATA, OR FOR DIRECT, INDIRECT, SPECIAL, CONSEQUENTIAL,
 * INCIDENTAL OR PUNITIVE DAMAGES, HOWEVER CAUSED AND REGARDLESS OF THE THEORY
 * OF LIABILITY, ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE,
 * EVEN IF SUN HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 *  
 * You acknowledge that this software is not designed, licensed or intended
 * for use in the design, construction, operation or maintenance of any
 * nuclear facility. 
 */

/**
* Plot represents the state of an entity at a point in time,
* including velocity, and desired velocity (which might be 
* the same the current velocity), and status.
*/
public class Plot
{
    /** The default plot type. */
    public static final byte TYPE_PLOT = (byte) 0;
    
    /** Signifies that entity has exited the space. */
    public static final byte TYPE_EXIT = (byte) 128;
    
    // event data
    public byte type;
    public int tick;

    // state
    public byte damage;
    public byte front;
    public byte left;
    public byte right;
    public byte rear;
    public byte torpA;
    public byte torpB;

    // position 
    public int x;
    public int y;
    public short dir;
    public short mag;

    // course
    public short tdir;
    public short tmag;
    public long target;
    
    /**
    * Default constructor: tick is zero and all elements are NIL.
    */
    public Plot()
    {
        tick = 0;
        x = y = dir = mag = tdir = tmag = Entity.NIL;
    }
    
    /**
    * Copy constructor: creates complete copy.
    */
    public Plot( Plot inPlot )
    {
        this();
        
        tick = inPlot.tick;
        x = inPlot.x;
        y = inPlot.y;
        dir = inPlot.dir;
        mag = inPlot.mag;
        tdir = inPlot.tdir;
        tmag = inPlot.tmag;
        
        damage = inPlot.damage;
        front = inPlot.front;
        left = inPlot.left;
        right = inPlot.right;
        rear = inPlot.rear;
        torpA = inPlot.torpA;
        torpB = inPlot.torpB;
        
        type = inPlot.type;
        target = inPlot.target;
    }
    
    /**
    * Constructor creates instance from serialized byte array.
    */
    public Plot( byte[] bytes )
    {
        this();
        fromBytes( bytes );
    }
    
    public String toString()
    {
        return "[PLOT:"+tick+"::"+x+":"+y+"::"+dir+":"+mag+"::"+tdir+":"+tmag+":"+Long.toString( target )+"]";
    }
    
    public byte[] toBytes()
    {
        byte[] bytes = new byte[44];
        int i = 0;
 
        bytes[i++] = type;

        bytes[i++] = (byte) ( 0xff & ( tick >> 24 ) );
        bytes[i++] = (byte) ( 0xff & ( tick >> 16 ) );
        bytes[i++] = (byte) ( 0xff & ( tick >> 8 ) );
        bytes[i++] = (byte) ( 0xff & ( tick ) );
        
        bytes[i++] = damage;

        bytes[i++] = front;
        bytes[i++] = left;
        bytes[i++] = right;
        bytes[i++] = rear;

        bytes[i++] = torpA;
        bytes[i++] = torpB;
        
        bytes[i++] = (byte) ( 0xff & ( x >> 24 ) );
        bytes[i++] = (byte) ( 0xff & ( x >> 16 ) );
        bytes[i++] = (byte) ( 0xff & ( x >> 8 ) );
        bytes[i++] = (byte) ( 0xff & ( x ) );

        bytes[i++] = (byte) ( 0xff & ( y >> 24 ) );
        bytes[i++] = (byte) ( 0xff & ( y >> 16 ) );
        bytes[i++] = (byte) ( 0xff & ( y >> 8 ) );
        bytes[i++] = (byte) ( 0xff & ( y ) );
        
        bytes[i++] = (byte) ( 0xff & ( dir >> 24 ) );
        bytes[i++] = (byte) ( 0xff & ( dir >> 16 ) );
        bytes[i++] = (byte) ( 0xff & ( dir >> 8 ) );
        bytes[i++] = (byte) ( 0xff & ( dir ) );
        
        bytes[i++] = (byte) ( 0xff & ( mag >> 24 ) );
        bytes[i++] = (byte) ( 0xff & ( mag >> 16 ) );
        bytes[i++] = (byte) ( 0xff & ( mag >> 8 ) );
        bytes[i++] = (byte) ( 0xff & ( mag ) );
        
        bytes[i++] = (byte) ( 0xff & ( tdir >> 24 ) );
        bytes[i++] = (byte) ( 0xff & ( tdir >> 16 ) );
        bytes[i++] = (byte) ( 0xff & ( tdir >> 8 ) );
        bytes[i++] = (byte) ( 0xff & ( tdir ) );
        
        bytes[i++] = (byte) ( 0xff & ( tmag >> 24 ) );
        bytes[i++] = (byte) ( 0xff & ( tmag >> 16 ) );
        bytes[i++] = (byte) ( 0xff & ( tmag >> 8 ) );
        bytes[i++] = (byte) ( 0xff & ( tmag ) );

        bytes[i++] = (byte) ( 0xff & ( target >> 56 ) );
        bytes[i++] = (byte) ( 0xff & ( target >> 48 ) );
        bytes[i++] = (byte) ( 0xff & ( target >> 40 ) );
        bytes[i++] = (byte) ( 0xff & ( target >> 32 ) );
        bytes[i++] = (byte) ( 0xff & ( target >> 24 ) );
        bytes[i++] = (byte) ( 0xff & ( target >> 16 ) );
        bytes[i++] = (byte) ( 0xff & ( target >> 8 ) );
        bytes[i++] = (byte) ( 0xff & ( target ) );
        
        return bytes;
    }
    
    public void fromBytes( byte[] bytes )
    {
        int i = 0;
    
        type = bytes[i++];

        tick  = 0;
        tick += ( 0xff & bytes[i++] ) << 24;
        tick += ( 0xff & bytes[i++] ) << 16;
        tick += ( 0xff & bytes[i++] ) << 8;
        tick += ( 0xff & bytes[i++] );

        damage = bytes[i++];

        front = bytes[i++];
        left = bytes[i++];
        right = bytes[i++];
        rear = bytes[i++];

        torpA = bytes[i++];
        torpB = bytes[i++];

        x  = 0;
        x += ( 0xff & bytes[i++] ) << 24;
        x += ( 0xff & bytes[i++] ) << 16;
        x += ( 0xff & bytes[i++] ) << 8;
        x += ( 0xff & bytes[i++] );

        y  = 0;
        y += ( 0xff & bytes[i++] ) << 24;
        y += ( 0xff & bytes[i++] ) << 16;
        y += ( 0xff & bytes[i++] ) << 8;
        y += ( 0xff & bytes[i++] );
        
        dir  = 0;
        dir += ( 0xff & bytes[i++] ) << 24;
        dir += ( 0xff & bytes[i++] ) << 16;
        dir += ( 0xff & bytes[i++] ) << 8;
        dir += ( 0xff & bytes[i++] );

        mag  = 0;
        mag += ( 0xff & bytes[i++] ) << 24;
        mag += ( 0xff & bytes[i++] ) << 16;
        mag += ( 0xff & bytes[i++] ) << 8;
        mag += ( 0xff & bytes[i++] );
        
        tdir  = 0;
        tdir += ( 0xff & bytes[i++] ) << 24;
        tdir += ( 0xff & bytes[i++] ) << 16;
        tdir += ( 0xff & bytes[i++] ) << 8;
        tdir += ( 0xff & bytes[i++] );

        tmag  = 0;
        tmag += ( 0xff & bytes[i++] ) << 24;
        tmag += ( 0xff & bytes[i++] ) << 16;
        tmag += ( 0xff & bytes[i++] ) << 8;
        tmag += ( 0xff & bytes[i++] );
        
        target  = 0;
        target += (long) ( 0xff & bytes[i++] ) << 56;
        target += (long) ( 0xff & bytes[i++] ) << 48;
        target += (long) ( 0xff & bytes[i++] ) << 40;
        target += (long) ( 0xff & bytes[i++] ) << 32;
        target += (long) ( 0xff & bytes[i++] ) << 24;
        target += (long) ( 0xff & bytes[i++] ) << 16;
        target += (long) ( 0xff & bytes[i++] ) << 8;
        target += (long) ( 0xff & bytes[i++] );
    }
    
    public static void main( String[] argv )
    {
        Plot plot = new Plot();
        plot.tick = Integer.MAX_VALUE-1;
        plot.x = Integer.MAX_VALUE-1;
        plot.y = Integer.MIN_VALUE+1;
        plot.dir = Short.MAX_VALUE-1;
        plot.mag = Short.MIN_VALUE+1;
        plot.tdir = Short.MAX_VALUE-1;
        plot.tmag = Short.MIN_VALUE+1;
        plot.target = Long.MAX_VALUE-1;
        System.out.println( plot );
        System.out.println();
        
        byte[] bytes = plot.toBytes();
        //System.out.println( "bytes: " );
        //for ( int i = 0; i < bytes.length; i++ ) System.out.println( i + " : " + bytes[i] );
        //System.out.println();
        
        System.out.println( "result: " );
        Plot result = new Plot();
        result.fromBytes( bytes );
        System.out.println( result );
        System.out.println( Integer.toString( result.tick ) );
        System.out.println( Integer.toString( result.x ) );
        System.out.println( Integer.toString( result.y ) );
        System.out.println( Integer.toString( result.dir ) );
        System.out.println( Integer.toString( result.mag ) );
        System.out.println( Integer.toString( result.tdir ) );
        System.out.println( Integer.toString( result.tmag ) );
        System.out.println( Long.toString( result.target ) );
        System.out.println();

        System.out.println( "reuse: " );
        plot.fromBytes( bytes );
        System.out.println( plot );
        System.out.println( Integer.toString( plot.tick ) );
        System.out.println( Integer.toString( plot.x ) );
        System.out.println( Integer.toString( plot.y ) );
        System.out.println( Integer.toString( plot.dir ) );
        System.out.println( Integer.toString( plot.mag ) );
        System.out.println( Integer.toString( plot.tdir ) );
        System.out.println( Integer.toString( plot.tmag ) );
        System.out.println( Long.toString( plot.target ) );
        System.out.println();
    }
}

