/* License
 * 
 * Copyright 1994-2006 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *  
 *  * Redistribution of source code must retain the above copyright notice,
 *	this list of conditions and the following disclaimer.
 * 
 *  * Redistribution in binary form must reproduce the above copyright notice,
 *	this list of conditions and the following disclaimer in the
 *	documentation and/or other materials provided with the distribution.
 * 
 * Neither the name of Sun Microsystems, Inc. or the names of contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *  
 * This software is provided "AS IS," without a warranty of any kind. ALL
 * EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND WARRANTIES, INCLUDING
 * ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE
 * OR NON-INFRINGEMENT, ARE HEREBY EXCLUDED. SUN MICROSYSTEMS, INC. ("SUN")
 * AND ITS LICENSORS SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY LICENSEE
 * AS A RESULT OF USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE OR ITS
 * DERIVATIVES. IN NO EVENT WILL SUN OR ITS LICENSORS BE LIABLE FOR ANY LOST
 * REVENUE, PROFIT OR DATA, OR FOR DIRECT, INDIRECT, SPECIAL, CONSEQUENTIAL,
 * INCIDENTAL OR PUNITIVE DAMAGES, HOWEVER CAUSED AND REGARDLESS OF THE THEORY
 * OF LIABILITY, ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE,
 * EVEN IF SUN HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 *  
 * You acknowledge that this software is not designed, licensed or intended
 * for use in the design, construction, operation or maintenance of any
 * nuclear facility. 
 */

import javax.microedition.lcdui.Display;
import javax.microedition.io.*;
import java.io.*;
import java.util.*;

/**
* Synchronizes a Space with a remote server.
*/
public class Synchronizer implements Runnable
{
    /**
    * The midlet will poll no more frequently than the
    * specified number of milliseconds.
    */
    public final static int POLL_INTERVAL = 1000;
    
    // the last known time on the server
    private int tick;

    // the controlling midlet
    private SpaceWar midlet;
    
    // the space to synchronize
    private Space space;
    
    // the server url
    private String serverURL;
 
    // polling thread for model updates
    private Thread pollThread;
    private boolean isPolling;
    private long lastPollTime;
    private int lastToken;
    
    public Synchronizer( SpaceWar inMIDlet, Space inSpace, String inURL )
    {
        midlet = inMIDlet;
        space = inSpace;
        serverURL = inURL;

        pollThread = new Thread( this );
        lastPollTime = 0;
        lastToken = 0;
    }
    
    /**
    * Starts synchronizing the space.
    */
    public void connect()
    {
        isPolling = true;
        pollThread.start();
    }
    
    /**
    * Stops synchronizing the space.
    */
    public void disconnect()
    {
        isPolling = false; // exits thread
    }
    
    /**
    * Returns the last known server time.
    */
    public int getTick()
    {
        return tick;
    }
    
    /**
    * Polls the server, POSTing any pending events,
    * otherwise GETing the events since the last poll.
    * Returns the latest timestamp from the server.
    */ 
    private int poll( int lastToken ) throws IOException
    {
        int result = lastToken;
        HttpConnection connection = null;
        DataInputStream input = null;
        DataOutputStream output = null;

        try 
        {
            String url = serverURL + "?token=" + lastToken;
            connection = (HttpConnection) Connector.open( url );
            Vector eventsToSend = null;
    
            // if items are in the queue to be POSTed
            if ( !space.pendingEvents.isEmpty() )
            {
                // assemble the request data in the server's specified format
                connection.setRequestMethod(HttpConnection.POST);
                output = new DataOutputStream( connection.openOutputStream() );
                
                // removal needs to be an atomic operation
                synchronized ( space.pendingEvents )
                {
                    eventsToSend = space.pendingEvents;
                    space.pendingEvents = new Vector();
                }
                
                byte[] bytes;
                Long id;
                Plot plot;
                Enumeration e = eventsToSend.elements();
                output.writeInt( tick ); // current time
                output.writeInt( lastToken ); // current state
                output.writeShort( eventsToSend.size() / 2 ); // entity count
                while ( e.hasMoreElements() )
                {
                    id = (Long) e.nextElement();
                    plot = (Plot) e.nextElement();
                    // System.out.println( "send plot: " + id + " : " + plot );            
                    bytes = plot.toBytes();
                    output.writeLong( id.longValue() );
                    output.writeByte( (byte) bytes.length ); // event length
                    output.write( bytes );
                }
                output.flush();
            }
            else // just GET the latest events
            {
                connection.setRequestMethod(HttpConnection.GET);
            }

            if ( connection.getResponseCode() != HttpConnection.HTTP_OK ) 
            {
                // if we failed to send events
                if ( eventsToSend != null )
                { // put them back on the queue
                    synchronized ( space.pendingEvents )
                    { // copy anything that might have happened during send
                        Enumeration e = space.pendingEvents.elements();
                        while ( e.hasMoreElements() ) eventsToSend.addElement( e.nextElement() );
                        space.pendingEvents = eventsToSend; // revert
                    }
                }
                throw new IOException("HTTP response code: " 
                    + connection.getResponseCode() + " : " + connection.getResponseMessage() );
            }
            
            // store results in a data structure for use on event thread
            final Hashtable idToEvents = new Hashtable();
            Vector history;
            byte[] event;

            // disassemble the response data from the server's specified format
            Long id;
            input = new DataInputStream( connection.openInputStream() );
            tick = input.readInt(); // server's current time
            result = input.readInt(); // return the next sync token

            int eventSize;
            short eventCount = input.readShort(); // entity count
            for ( int i = 0; i < eventCount; i++ )
            {
                id = new Long( input.readLong() ); // entity id
                // System.out.print( id + " " );
                history = (Vector) idToEvents.get( id );
                if ( history == null )
                {
                    history = new Vector();
                    idToEvents.put( id, history );
                }
                
                eventSize = input.readByte(); // event size
                // System.out.print( eventSize + " " );
                event = new byte[eventSize];
                input.readFully( event ); // event data
                history.addElement( new Plot( event ) );
                // System.out.println();
            }
                // System.out.println();
            
            // synchronize the model update on the event thread
            Display.getDisplay( midlet ).callSerially( new Runnable() 
            {
                public void run()
                {
                    Long id;
                    Entity entity;
                    Enumeration keys, events;
                    keys = idToEvents.keys();
                    while ( keys.hasMoreElements() )
                    {
                        id = (Long) keys.nextElement();
                        events = ((Vector)idToEvents.get( id )).elements();
                        while ( events.hasMoreElements() )
                        {
                            space.insertPlotForID( (Plot) events.nextElement(), id.longValue() );
                        }
                    }
                    
                    midlet.playerEntity = midlet.getEntityForPlayer( SpaceWar.playerID );
                    
                    // if no ship found: create a new ship
                    if ( midlet.playerEntity == null )
                    {
                        midlet.playerEntity = new Ship();
                    
                        // pick a random starting position
                        Random random = new Random();
                        Plot plot = new Plot();
                        plot.x = 500000 + ( random.nextInt() % 25000 );
                        plot.y = 500000 + ( random.nextInt() % 25000 );
                        plot.dir = (short) 0;
                        plot.mag = (short) 0;
                        
                        // add to space
                        space.addEntity( midlet.playerEntity );
                        
                        // post enter event
                        space.postEventForID( plot, midlet.playerEntity.getID() );
                    }
                }
            } );
            
        } 
        catch (ClassCastException e) 
        {
            throw new IllegalArgumentException( "Not an HTTP URL" );
        }
        finally 
        {
            if (input != null)
                input.close();
            if (output != null)
                output.close();
            if (connection != null)
                connection.close();
        }
        return result;
    }
    
    // interface Runnable: the poll thread
    
    public void run()
    {
        long diff;
        while ( isPolling )
        {
            diff = System.currentTimeMillis() - lastPollTime;
            if ( diff > POLL_INTERVAL )
            {
                //System.out.println( "polling: since last poll: " + diff );            
                lastPollTime = System.currentTimeMillis();
                try
                {
                    lastToken = poll( lastToken );                    
                }
                catch ( IOException ioe )
                {
                    System.err.println( "Error while polling: " + ioe );
                    ioe.printStackTrace();
                }
            }
            else
            {
                try
                {
                    Thread.sleep( POLL_INTERVAL - diff );
                } 
                catch ( InterruptedException exc )
                {
                    // resume
                }
            }
        }
    }
    
    
}
