/* License
 * 
 * Copyright 1994-2006 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *  
 *  * Redistribution of source code must retain the above copyright notice,
 *	this list of conditions and the following disclaimer.
 * 
 *  * Redistribution in binary form must reproduce the above copyright notice,
 *	this list of conditions and the following disclaimer in the
 *	documentation and/or other materials provided with the distribution.
 * 
 * Neither the name of Sun Microsystems, Inc. or the names of contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *  
 * This software is provided "AS IS," without a warranty of any kind. ALL
 * EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND WARRANTIES, INCLUDING
 * ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE
 * OR NON-INFRINGEMENT, ARE HEREBY EXCLUDED. SUN MICROSYSTEMS, INC. ("SUN")
 * AND ITS LICENSORS SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY LICENSEE
 * AS A RESULT OF USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE OR ITS
 * DERIVATIVES. IN NO EVENT WILL SUN OR ITS LICENSORS BE LIABLE FOR ANY LOST
 * REVENUE, PROFIT OR DATA, OR FOR DIRECT, INDIRECT, SPECIAL, CONSEQUENTIAL,
 * INCIDENTAL OR PUNITIVE DAMAGES, HOWEVER CAUSED AND REGARDLESS OF THE THEORY
 * OF LIABILITY, ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE,
 * EVEN IF SUN HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 *  
 * You acknowledge that this software is not designed, licensed or intended
 * for use in the design, construction, operation or maintenance of any
 * nuclear facility. 
 */

import javax.microedition.lcdui.*;
import javax.microedition.lcdui.game.*;

/**
* A sprite that corresponds to an entity.
*/
public class EntitySprite extends Sprite
{
    private final static int DIR_INCREMENT = 10;
    private final static int MAG_INCREMENT = 100;

    // the entity this sprite will render
    private Entity entity;

    // the user's target direction and magnitude
    private short tdir, tmag;
    
    // the last known plot
    private Plot plot;
    
    public EntitySprite( Entity inEntity, int tick, Image image, int frameWidth, int frameHeight )
    {
        super(image, frameWidth, frameHeight);
        defineReferencePixel(frameWidth / 2, frameHeight / 2);

        entity = inEntity;
        tdir = tmag = Entity.NIL;
        
        // update target trajectory to match current plot
        Plot plot = entity.getPositionForTick( tick );
        if ( plot != null)
        {
            if ( plot.tdir != Entity.NIL ) tdir = plot.tdir;
            if ( plot.tmag != Entity.NIL ) tmag = plot.tmag;
            updatePosition( plot );
        }
    }
    
    /**
    * Applies the specified delta values to the 
    * relative direction and magnitude.
    */
    public void adjustTrajectory( int dir, int mag, int tick )
    {
        tdir = (short) ((tdir+dir)%360);
        tmag += (short) mag;
        tmag = (short) Math.max( Math.min( 
            tmag, entity.getMagnitudeMax(tick) ), entity.getMagnitudeMin(tick) );
        //System.out.println( "adjustTrajectory: " + tdir + " : " + tmag );        
    }
    
    /**
    * Called to update the frame before we're told to paint.
    */
    public void updatePosition( int tick )
    {
        updatePosition( entity.getPositionForTick( tick ) );
    }
    
    /**
    * Called to update the frame before we're told to paint.
    */
    public void updatePosition( Plot inPlot )
    {
        plot = inPlot;
        if ( plot == null || ( plot.type & Plot.TYPE_EXIT ) != 0 )
        { // not yet on board, or has exited board
            setVisible( false );
            return;
        }
        
        setVisible( true );
        setPosition( 
            (int) plot.x/SpaceCanvas.CLICKS_PER_PIXEL, 
            (int) plot.y/SpaceCanvas.CLICKS_PER_PIXEL ); 
        
        int frame = plot.dir; // ship facing
        setFrame( (frame%90)/10 );
        
        switch ( frame/90 )
        {
            case 0: setTransform( TRANS_ROT270 ); break;
            case 1: setTransform( TRANS_NONE );   break;
            case 2: setTransform( TRANS_ROT90 );  break;
            case 3: setTransform( TRANS_ROT180 ); break;
        }
        //System.out.println( "readFromEntity: " + this + " : " + plot.x + " : " + plot.y );
    }
    
    /**
    * Returns the user's speed and directional commands
    * as a plot, null if the plot is unchanged from current.
    * The tick of the returned plot is NIL.
    */
    public Plot getCurrentPlot( int tick )
    {
        Plot p = entity.getPositionForTick( tick );
        if ( p == null || ( p.mag == tmag && p.dir == tdir ) ) return null;

        // plot desired course, if any
        Plot result = new Plot( p );
        result.tick = tick;
        result.tdir = (short) ((tdir+360)%360); // positive required
        result.tmag = tmag;
        
        return result;
    }
    
    // rendering
    
    /**
    * Draws 'heads-up display' trajectory data.
    * Ordinarily, would have liked to just subclass 
    * paint, but paint is final.
    */ 
    public void paintHUD( Graphics g, int x, int y )
    {
        int diff;
        int dir = plot.dir;
        
        // draw turn signal
        if ( tdir != Entity.NIL )
        {
            diff = dir-tdir;
            if ( diff < -180 ) diff += 360; else if ( diff > 180 ) diff -= 360;
            
            g.setColor( 64, 64, 255 );
            g.drawArc( x-10, y-10, getWidth()+20, getHeight()+20, (360-dir)+90, diff );
        }
        
        // draw accel/decel
        if ( tmag != Entity.NIL )
        {
            diff = ( tmag - plot.mag );
            if ( diff > 0 )
            {
                g.setColor( 0, 128, 0 );
                for ( int i = 0; i < diff; i++ )
                {
                    g.drawArc( x-12-2*i, y-12-2*i, 
                        getWidth() + 24+4*i, getHeight() + 24+4*i, 
                            (360-dir)+90-15, 30 );
                }
            }
            else
            if ( diff < 0 )
            {
                g.setColor( 128, 0, 0 );
                for ( int i = 0; i < -diff; i++ )
                {
                    g.drawArc( x-12-2*i, y-12-2*i, 
                        getWidth() + 24+4*i, getHeight() + 24+4*i, 
                            (360-dir)+90+180-15, 30 );
                }
            }
        }
        
        // draw shields
        g.setColor( 128, 128, 128 );
        drawHUDArc( g, x, y, dir, 3, 45, 80 ); 
        drawHUDArc( g, x, y, dir, 2, 45, 80 ); 
        drawHUDArc( g, x, y, dir, 1, 45, 80 ); 

        drawHUDArc( g, x, y, dir, 2, 135, 80 ); 
        drawHUDArc( g, x, y, dir, 1, 135, 80 ); 
        
        drawHUDArc( g, x, y, dir, 1, 225, 80 ); 

        drawHUDArc( g, x, y, dir, 2, 315, 80 ); 
        drawHUDArc( g, x, y, dir, 1, 315, 80 ); 
    }
    
    private void drawHUDArc( Graphics g, int x, int y, int dir, int dist, int start, int length )
    {
        g.drawArc( 
            x-dist*2, y-dist*2, 
            getWidth()+dist*4, getHeight()+dist*4, 
            (360-dir)+90+5 - start, length );
    }
}
