/* License
 * 
 * Copyright 1994-2006 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *  
 *  * Redistribution of source code must retain the above copyright notice,
 *	this list of conditions and the following disclaimer.
 * 
 *  * Redistribution in binary form must reproduce the above copyright notice,
 *	this list of conditions and the following disclaimer in the
 *	documentation and/or other materials provided with the distribution.
 * 
 * Neither the name of Sun Microsystems, Inc. or the names of contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *  
 * This software is provided "AS IS," without a warranty of any kind. ALL
 * EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND WARRANTIES, INCLUDING
 * ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE
 * OR NON-INFRINGEMENT, ARE HEREBY EXCLUDED. SUN MICROSYSTEMS, INC. ("SUN")
 * AND ITS LICENSORS SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY LICENSEE
 * AS A RESULT OF USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE OR ITS
 * DERIVATIVES. IN NO EVENT WILL SUN OR ITS LICENSORS BE LIABLE FOR ANY LOST
 * REVENUE, PROFIT OR DATA, OR FOR DIRECT, INDIRECT, SPECIAL, CONSEQUENTIAL,
 * INCIDENTAL OR PUNITIVE DAMAGES, HOWEVER CAUSED AND REGARDLESS OF THE THEORY
 * OF LIABILITY, ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE,
 * EVEN IF SUN HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 *  
 * You acknowledge that this software is not designed, licensed or intended
 * for use in the design, construction, operation or maintenance of any
 * nuclear facility. 
 */

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

/**
* Space is a container for Entities.
* The entities are located within a 2D-space with a
* coordinate system ranging from Integer.MIN_VALUE and
* Integer.MAX_VALUE (-2,147,483,648 to 2,147,483,647).
*/
public class Space 
{
    /** Known entities. */
    private Hashtable entities;
    
    /**
    * out-going event queue (FIFO).
    * elements alternate between entity id and event object.
    * package access: used by synchronizer. */
    Vector pendingEvents;
    
    public Space()
    {
        entities = new Hashtable();
        pendingEvents = new Vector();
    }
    
    /**
    * Records the specified event for the specified id.
    * If the entity does not exist, it will be created.
    */
    public void insertPlotForID( Plot inPlot, long inID )
    { // System.out.println( "insertPlotForID: " + inPlot + " : " + Long.toString( inID ) );    
        Long id = new Long( inID );
        Entity entity = (Entity) entities.get( id );
        
        if ( entity == null && ( inPlot.type & Plot.TYPE_EXIT ) == 0 )
        { // register new object
            if ( ( inID & Entity.TYPE_SHIP ) != 0 )
            {
                entity = new Ship( inID ); 
            }
            else
            if ( ( inID & Entity.TYPE_SHOT ) != 0 )
            {
                entity = new Shot( inID ); 
            }
            else
            {
                System.err.println( "Unrecognized type for id: " + inID );
                return;
            }
            addEntity( entity );
        }
        
        if ( entity != null )
        {
            entity.insertPlot( inPlot );
        }
    }
    
    /**
    * Retrieves the entity for the specified id,
    * or null if the id doesn't exist.
    */
    public void addEntity( Entity inEntity )
    {
        entities.put( new Long( inEntity.getID() ), inEntity );
    }
    
    /**
    * Retrieves the entity for the specified id,
    * or null if the id doesn't exist.
    */
    public Entity getEntityForID( long id )
    {
        return (Entity) entities.get( new Long( id ) );
    }
    
    /**
    * Obtains an enumerator over the entities.
    */
    public Enumeration entities()
    {
        return entities.elements();
    }
    
    /**
    * Posts an event in the queue for transmitting to server
    * for the entity with the specified id.
    */
    public void postEventForID( Plot inPlot, long id )
    {
        // adding two elements must be atomic operation
        synchronized ( pendingEvents )
        { 
            pendingEvents.addElement( new Long( id ) );
            pendingEvents.addElement( inPlot );
        }
        insertPlotForID( inPlot, id );
    }
    
    /**
    * Removes all processed events and exited entities as of specified tick.
    */
    public void compact( int tick )
    {
        Entity entity;
        Enumeration e = entities();
        while ( e.hasMoreElements() )
        {
            entity = (Entity) e.nextElement();
            if ( !entity.isAlive( tick ) )
            {
                entities.remove( entity );
            }
        }
    }
    
}
