/* License
 * 
 * Copyright 1994-2006 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *  
 *  * Redistribution of source code must retain the above copyright notice,
 *	this list of conditions and the following disclaimer.
 * 
 *  * Redistribution in binary form must reproduce the above copyright notice,
 *	this list of conditions and the following disclaimer in the
 *	documentation and/or other materials provided with the distribution.
 * 
 * Neither the name of Sun Microsystems, Inc. or the names of contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *  
 * This software is provided "AS IS," without a warranty of any kind. ALL
 * EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND WARRANTIES, INCLUDING
 * ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE
 * OR NON-INFRINGEMENT, ARE HEREBY EXCLUDED. SUN MICROSYSTEMS, INC. ("SUN")
 * AND ITS LICENSORS SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY LICENSEE
 * AS A RESULT OF USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE OR ITS
 * DERIVATIVES. IN NO EVENT WILL SUN OR ITS LICENSORS BE LIABLE FOR ANY LOST
 * REVENUE, PROFIT OR DATA, OR FOR DIRECT, INDIRECT, SPECIAL, CONSEQUENTIAL,
 * INCIDENTAL OR PUNITIVE DAMAGES, HOWEVER CAUSED AND REGARDLESS OF THE THEORY
 * OF LIABILITY, ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE,
 * EVEN IF SUN HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 *  
 * You acknowledge that this software is not designed, licensed or intended
 * for use in the design, construction, operation or maintenance of any
 * nuclear facility. 
 */

import javax.microedition.lcdui.*;
import javax.microedition.lcdui.game.*;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;
import java.util.Timer;
import java.util.TimerTask;
import java.io.IOException;

/**
* SpaceCanvas coordinates its sprites with the entities in a Space.
*/
public class SpaceCanvas extends GameCanvas 
{
    /** Determines the number of frames per second. */
    public final static int FRAMES_PER_SECOND = 10;

    /** Determines the mapping of distance
    *   units to screen pixels. */ 
    public final static int CLICKS_PER_PIXEL = 500;
    
    /** The delay before commiting the focused entity
    *   with the new course. */
    public final static int INPUT_DELAY = 500;
    
    private Space space;
    private Hashtable entityToSprite;
    private LayerManager layerManager;
    
    // tracks simulation's current time
    private long timebase;
    private int tick;

    // input coalescing: to avoid immediately plotting each keypress
    private Timer inputTimer;
    private TimerTask inputTask;
    
    // custom background layer
    private Image bgimage;
    int bgwidth, bgheight;
    
    public SpaceCanvas( Space inSpace )
    {
        super( false ); // we're not using key states
        space = inSpace;
        entityToSprite = new Hashtable();
        layerManager = new LayerManager();

        syncTick( 0 ); // default to tick zero
        
        inputTimer = new Timer();
        
        // load background tile if any
        bgimage = null;
        try
        {
            bgimage = Image.createImage( "/space-tile.jpg" );
            bgwidth = bgimage.getWidth();
            bgheight = bgimage.getHeight();
        } 
        catch ( IOException ioe )
        {
            System.out.println( "No background found: resuming" );
            ioe.printStackTrace();
        }
    }
    
    /**
    * Gets the current tick that will be rendered.
    */
    public int getTick()
    {
        return (int) ( tick + ( ( System.currentTimeMillis() - timebase ) / SpaceWar.MILLIS_PER_TICK ) );
    }
    
    /**
    * Synchronizes the "start tick" for the canvas.
    * The canvas will render the real-time difference between
    * the time that this method was called and the time of
    * rendering relative to the specified tick.
    */
    public void syncTick( int inTick )
    {
        // System.out.println( "syncTick: " + inTick + " : diff = " + (inTick - tick) );  
        tick = inTick;
        timebase = System.currentTimeMillis();
    }
        
    /**
    * Creates a sprite that corresponds to the specified entity.
    */
    private EntitySprite createSprite( Entity inEntity, String inImageResource ) 
    {
        try
        {
            Image image = Image.createImage( inImageResource );
            EntitySprite result = new EntitySprite( 
                inEntity, getTick(), image, image.getHeight(), image.getHeight() );
            return result;
        } 
        catch ( IOException ioe )
        {
            ioe.printStackTrace();
        }
        return null;
    }
    
    /**
    * Called by paint() to update the position of each of the sprites.
    */
    private void syncSprites()
    {
        Enumeration e;
        
        // initialize discard list
        Vector sprites = new Vector();
        e = entityToSprite.elements();
        while ( e.hasMoreElements() )
        {
            sprites.addElement( e.nextElement() );
        }
        
        // System.out.println( "syncSprites: " + sprites );

        Entity entity;
        EntitySprite sprite;
        e = space.entities();
        while ( e.hasMoreElements() )
        {
            entity = (Entity) e.nextElement();
            sprite = (EntitySprite) entityToSprite.get( entity );

            // System.out.println( "syncSprites: " + entity + " : " + sprite );
            
            if ( sprite == null )
            { // create sprite and add it to layer manager
                sprite = createSprite( entity, entity.getImageName() );
                layerManager.insert( sprite, 0 );
                entityToSprite.put( entity, sprite );
            }
            else
            { // remove from discard list
                sprites.removeElement( sprite );
            }
            
            // update sprite with latest data
            sprite.updatePosition( getTick() );
        }
        
        e = sprites.elements();
        while ( e.hasMoreElements() )
        {
            layerManager.remove( (Sprite) e.nextElement() );
        }
    }
    
    /**
    * Copies the user's speed and directional commands
    * into a plot placed on the entity's queue.
    */
    private void doPlot()
    {
        Entity focusedEntity = SpaceWar.playerEntity;
        EntitySprite sprite = (EntitySprite) entityToSprite.get( focusedEntity );
        if ( sprite != null )
        {
            Plot plot = sprite.getCurrentPlot( getTick()+10 );
            if ( plot != null )
            {
                space.postEventForID( plot, focusedEntity.getID() );
            }
        }
    }
    
    // rendering
    
    public void paint( Graphics g )
    {
        EntitySprite sprite = null;
        Entity focusedEntity = SpaceWar.playerEntity;
        if ( focusedEntity != null )
        {
            sprite = (EntitySprite) entityToSprite.get( focusedEntity );
        }

        // update the model
        if ( SpaceWar.instance.synchronizer != null )
        {
            int serverTime = SpaceWar.instance.synchronizer.getTick();
            if ( serverTime > getTick() )
            { 
                // server is ahead: catch up
                syncTick( serverTime );
            }
        }
        
        // update the sprite positions
        syncSprites();
        
        int width = getWidth();
        int height = getHeight();
        int viewx = 0;
        int viewy = 0;
        
        if ( sprite != null )
        {
            // render with the focused sprite centered
            viewx = -width/2+sprite.getX()+sprite.getWidth()/2;
            viewy = -height/2+sprite.getY()+sprite.getHeight()/2;
            layerManager.setViewWindow( 
                viewx, viewy, 
                width, height );
        }
        else
        {
            // render the "origin" region
            layerManager.setViewWindow( 0, 0, width, height );
        }

        // paint background
        if ( bgimage != null )
        {
            for ( int x = -(viewx%bgwidth)-bgwidth; x < width; x += bgwidth )
            {
                for ( int y = -(viewy%bgheight)-bgheight; y < height; y += bgheight )
                {
                    g.drawImage( bgimage, x, y, 0 );
                }
            }
        }
        else // no background tile: erase background
        {
            g.setColor( 0 );
            g.fillRect( 0, 0, width, height );
        }
        
        // paint sprites
        layerManager.paint( g, 0, 0 );
        
        // paint HUD
        if ( sprite != null )
        {
            sprite.paintHUD( g, 
                (getWidth()-sprite.getWidth())/2, 
                (getHeight()-sprite.getHeight())/2 );
        }
    }
    
    // input handling
    
    protected void keyPressed(int keyCode)
    { 
        EntitySprite sprite = null;
        Entity focusedEntity = SpaceWar.playerEntity;
        if ( focusedEntity != null )
        {
            sprite = (EntitySprite) entityToSprite.get( focusedEntity );
        }
        if ( sprite == null ) return;

        // convert game codes into appropriate keyCode as lastKeyPress
        int gameAction = getGameAction( keyCode );
        switch ( gameAction )
        {
            case UP:
                sprite.adjustTrajectory( 0, 1, getTick() );
                break;
            case DOWN:
                sprite.adjustTrajectory( 0, -1, getTick() );
                break;
            case LEFT:
                sprite.adjustTrajectory( -10, 0, getTick() );
                break;
            case RIGHT:
                sprite.adjustTrajectory( 10, 0, getTick() );
                break;
            case FIRE:
                onFireButton( focusedEntity );
                break;
            default:
                // leave unchanged
        }
        repaint();

        // reset the input timer
        if ( inputTask != null ) inputTask.cancel();
        inputTask = new TimerTask()
        {
            public void run()
            {   
                doPlot();
            }
        };
        inputTimer.schedule( inputTask, INPUT_DELAY );
    }
    
    protected void keyRepeated(int keyCode)
    { 
        keyPressed( keyCode );
    }
    
    protected void keyReleased(int keyCode)
    {
    }
    
    // launches a torpedo
    private void onFireButton( Entity inEntity )
    {
        Shot shot = new Shot();
        Plot plot = inEntity.getPositionForTick( getTick()+10 );
        plot.mag = (short) 7;
        
        // posting event will implictly create the entity on next poll
        space.postEventForID( plot, shot.getID() );

        plot = new Plot( plot );
        plot.type += Plot.TYPE_EXIT;
        plot.tick += 150;
        space.postEventForID( plot, shot.getID() );
    }
    
}
