/* License
 * 
 * Copyright 1994-2006 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *  
 *  * Redistribution of source code must retain the above copyright notice,
 *	this list of conditions and the following disclaimer.
 * 
 *  * Redistribution in binary form must reproduce the above copyright notice,
 *	this list of conditions and the following disclaimer in the
 *	documentation and/or other materials provided with the distribution.
 * 
 * Neither the name of Sun Microsystems, Inc. or the names of contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *  
 * This software is provided "AS IS," without a warranty of any kind. ALL
 * EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND WARRANTIES, INCLUDING
 * ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE
 * OR NON-INFRINGEMENT, ARE HEREBY EXCLUDED. SUN MICROSYSTEMS, INC. ("SUN")
 * AND ITS LICENSORS SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY LICENSEE
 * AS A RESULT OF USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE OR ITS
 * DERIVATIVES. IN NO EVENT WILL SUN OR ITS LICENSORS BE LIABLE FOR ANY LOST
 * REVENUE, PROFIT OR DATA, OR FOR DIRECT, INDIRECT, SPECIAL, CONSEQUENTIAL,
 * INCIDENTAL OR PUNITIVE DAMAGES, HOWEVER CAUSED AND REGARDLESS OF THE THEORY
 * OF LIABILITY, ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE,
 * EVEN IF SUN HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 *  
 * You acknowledge that this software is not designed, licensed or intended
 * for use in the design, construction, operation or maintenance of any
 * nuclear facility. 
 */

import javax.microedition.lcdui.*;
import javax.microedition.io.*;
import javax.microedition.midlet.*;
import javax.microedition.rms.*;
import java.io.*;
import java.util.*;

public class SpaceWar extends MIDlet implements CommandListener
{
    /**
    * Singleton reference to most recently created instance.
    */
    public static SpaceWar instance = null;

    /** 
    * Determines the mapping of realtime milliseconds 
    * to gametime "ticks".  Must be in sync with server!
    */
    public final static int MILLIS_PER_TICK = 100;
    
    /** 
    * The player's unique identifier.
    * This is a permanent id generated on first launch.
    */
    public static int playerID; 
    
    /** 
    * The player's ship, or null if no ship is currently controlled.
    */
    public static Entity playerEntity;

    // midp necessities   
    Display display;
    Command exitCmd;
    
    // the model
    Space space;
    
    // the view
    SpaceCanvas canvas;
    
    // the connection
    Synchronizer synchronizer;
    
    // timer thread for view repaints
    private Timer repaintTimer;
        
    public SpaceWar()
    {
        instance = this;
        space = new Space();
        canvas = new SpaceCanvas( space );
        playerEntity = null; // initialized on connect

        // connect to the server specified in the jad
        String host = getAppProperty( "Game-Server" );
        if ( host != null )
        {
            synchronizer = new Synchronizer( this, space, host );
        }
        else
        {
            System.out.println( "No Game-Server specified." );
        }
        
        // read player's persistent id
        if ( !readPrefs() )
        { 
            // doesn't exist: generate it
            playerID = (int) ( System.currentTimeMillis() >> 16 );
            playerID = (int) ( ( playerID << 16 ) + (short) new Random().nextInt() );
            // reasonable chance of uniqueness: time-based plus random.
            // you could get more sophisticated, if desired.
            
            // now persist it
            writePrefs();
        }
    }
	
    public void startApp()
    {
		if ( display == null )
		{
			display = Display.getDisplay( this );

            exitCmd = new Command( "Exit", Command.EXIT, 0 );
			canvas.addCommand( exitCmd );
			canvas.setCommandListener( this );
		}
        
		display.setCurrent( canvas );

        repaintTimer = new Timer();
        repaintTimer.schedule( new TimerTask()
        {
            public void run()
            {
                canvas.repaint();
            }
        }, 0, 1000/SpaceCanvas.FRAMES_PER_SECOND );
        
        if ( synchronizer != null ) 
        {
            synchronizer.connect();
        }
	}
	
    public void pauseApp()
    {
        repaintTimer.cancel();
        if ( synchronizer != null ) 
        {
            synchronizer.disconnect();
        }
    }
	
    public void destroyApp( boolean unconditional )
    {
        pauseApp();
    }
	
	public void commandAction( Command c, Displayable s )
    {   
        if ( c == exitCmd )
        {
            destroyApp( true );
            notifyDestroyed();
        }
	}
    
    /**
    * Retrieves an entity for the current player
    * or null if no entity currently belongs to the player.
    */
    public Entity getEntityForPlayer( int playerID )
    {
        Entity entity;
        Enumeration e = space.entities();
        while ( e.hasMoreElements() )
        {
            entity = (Entity) e.nextElement();
            if ( entity.isOwnedBy( playerID ) && ( entity.getID() & Entity.TYPE_SHIP ) != 0 )
            {
                return entity;
            }
        }
        return null;
    }
    
    /**
    * Read persistent preference data, if any.
    * Returns whether data was successfully read.
    */
    private boolean readPrefs()
    {
        try
        {
            boolean result = false;
            RecordStore store = 
                RecordStore.openRecordStore( "spacewar.prefs", true );
            
            try
            {
                byte[] bytes = store.getRecord( 1 );
                if ( bytes != null )
                {
                    DataInputStream input = new DataInputStream( 
                        new ByteArrayInputStream( bytes ) );
                    playerID = input.readInt(); // read the player's id
                    result = true;
                }
                else
                {
                    // no data found: do nothing
                }
            }
            catch ( InvalidRecordIDException notfound )
            {
                // do nothing
                System.err.println( "readPrefs: no existing record" );
            }
            catch ( IOException ioe )
            {
                // do nothing
                System.err.println( "readPrefs: error on read: " + ioe );
                ioe.printStackTrace();
            }
            
            store.closeRecordStore();
        }
        catch ( Throwable t )
        {
            // do nothing
            System.err.println( "readPrefs: unexpected error: " );
            t.printStackTrace();
        }
        return false;
    }
    
    /**
    * Writes persistent preference data.
    * Returns whether data was successfully written.
    */
    private boolean writePrefs()
    {
        try
        {
            RecordStore store = 
                RecordStore.openRecordStore( "spacewar.prefs", true );
            if ( store != null )
            {
                ByteArrayOutputStream bytes = new ByteArrayOutputStream();
                DataOutputStream output = new DataOutputStream( bytes );
                output.writeInt( playerID ); // write the player's id
                output.flush();
                
                byte[] data = bytes.toByteArray();
                if ( store.getNextRecordID() == 1 )
                {
                    store.addRecord( data, 0, data.length );
                }
                else
                {
                    store.setRecord( 1, data, 0, data.length );
                }
                
                store.closeRecordStore();
                return true;
            }
            else
            {
                System.err.println( "writePrefs: openRecordStore returned null" );
            }
        }
        catch ( Throwable t )
        {
            // do nothing
            System.err.println( "writePrefs: unexpected exception: " );
            t.printStackTrace();
        }
        return false;
    }
    
}
